@@ -16,7 +16,7 @@
 %                                 June 2000                                   %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2019 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright 1999-2020 ImageMagick Studio LLC, a non-profit organization      %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
@@ -38,32 +38,32 @@
 /*
   Include declarations.
 */
-#include "magick/studio.h"
-#include "magick/blob.h"
-#include "magick/blob-private.h"
-#include "magick/color-private.h"
-#include "magick/colormap.h"
-#include "magick/colormap-private.h"
-#include "magick/constitute.h"
-#include "magick/distort.h"
-#include "magick/exception.h"
-#include "magick/exception-private.h"
-#include "magick/cache.h"
-#include "magick/image.h"
-#include "magick/image-private.h"
-#include "magick/list.h"
-#include "magick/magic.h"
-#include "magick/magick.h"
-#include "magick/memory_.h"
-#include "magick/pixel-accessor.h"
-#include "magick/quantum-private.h"
-#include "magick/resource_.h"
-#include "magick/static.h"
-#include "magick/string_.h"
-#include "magick/module.h"
-#include "magick/transform.h"
-#include "magick/utility.h"
-#include "magick/utility-private.h"
+#include "MagickCore/studio.h"
+#include "MagickCore/blob.h"
+#include "MagickCore/blob-private.h"
+#include "MagickCore/color-private.h"
+#include "MagickCore/colormap.h"
+#include "MagickCore/colormap-private.h"
+#include "MagickCore/constitute.h"
+#include "MagickCore/exception.h"
+#include "MagickCore/exception-private.h"
+#include "MagickCore/cache.h"
+#include "MagickCore/distort.h"
+#include "MagickCore/image.h"
+#include "MagickCore/image-private.h"
+#include "MagickCore/list.h"
+#include "MagickCore/magic.h"
+#include "MagickCore/magick.h"
+#include "MagickCore/memory_.h"
+#include "MagickCore/resource_.h"
+#include "MagickCore/pixel-accessor.h"
+#include "MagickCore/quantum-private.h"
+#include "MagickCore/static.h"
+#include "MagickCore/string_.h"
+#include "MagickCore/module.h"
+#include "MagickCore/transform.h"
+#include "MagickCore/utility.h"
+#include "MagickCore/utility-private.h"
 
 typedef struct
    {
@@ -264,112 +264,103 @@ static void Rd_WP_DWORD(Image *image,size_t *d)
   return;
 }
 
-static MagickBooleanType InsertRow(unsigned char *p,ssize_t y,Image *image,
-  int bpp)
+static MagickBooleanType InsertRow(Image *image,unsigned char *p,ssize_t y,
+  int bpp,ExceptionInfo *exception)
 {
-  ExceptionInfo
-    *exception;
-
   int
     bit;
 
+  Quantum
+    index;
+
+  register Quantum
+    *q;
+
   ssize_t
     x;
 
-  register PixelPacket
-    *q;
-
-  IndexPacket
-    index;
-
-  register IndexPacket
-    *indexes;
-
-  exception=(&image->exception);
   q=QueueAuthenticPixels(image,0,y,image->columns,1,exception);
-  if (q == (PixelPacket *) NULL)
+  if (q == (Quantum *) NULL)
     return(MagickFalse);
-  indexes=GetAuthenticIndexQueue(image);
   switch (bpp)
     {
     case 1:  /* Convert bitmap scanline. */
       {
         for (x=0; x < ((ssize_t) image->columns-7); x+=8)
+        {
+          for (bit=0; bit < 8; bit++)
           {
-            for (bit=0; bit < 8; bit++)
-              {
-                index=((*p) & (0x80 >> bit) ? 0x01 : 0x00);
-                SetPixelIndex(indexes+x+bit,index);
-                if (index < image->colors)
-                  SetPixelRGBO(q,image->colormap+(ssize_t) index);
-                q++;
-              }
-            p++;
+            index=((*p) & (0x80 >> bit) ? 0x01 : 0x00);
+            SetPixelIndex(image,index,q);
+            if (index < image->colors)
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+            q+=GetPixelChannels(image);
           }
+          p++;
+        }
         if ((image->columns % 8) != 0)
           {
             for (bit=0; bit < (ssize_t) (image->columns % 8); bit++)
-              {
-                index=((*p) & (0x80 >> bit) ? 0x01 : 0x00);
-                SetPixelIndex(indexes+x+bit,index);
-                if (index < image->colors)
-                  SetPixelRGBO(q,image->colormap+(ssize_t) index);
-                q++;
-              }
+            {
+              index=((*p) & (0x80 >> bit) ? 0x01 : 0x00);
+              SetPixelIndex(image,index,q);
+              if (index < image->colors)
+                SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+              q+=GetPixelChannels(image);
+            }
             p++;
           }
         break;
       }
     case 2:  /* Convert PseudoColor scanline. */
       {
-        if ((image->storage_class != PseudoClass) ||
-            (indexes == (IndexPacket *) NULL))
-          break;
         for (x=0; x < ((ssize_t) image->columns-3); x+=4)
         {
-            index=ConstrainColormapIndex(image,(*p >> 6) & 0x3);
-            SetPixelIndex(indexes+x,index);
+            index=ConstrainColormapIndex(image,(*p >> 6) & 0x3,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
-            q++;
-            index=ConstrainColormapIndex(image,(*p >> 4) & 0x3);
-            SetPixelIndex(indexes+x,index);
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+            q+=GetPixelChannels(image);
+            index=ConstrainColormapIndex(image,(*p >> 4) & 0x3,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
-            q++;
-            index=ConstrainColormapIndex(image,(*p >> 2) & 0x3);
-            SetPixelIndex(indexes+x,index);
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+            q+=GetPixelChannels(image);
+            index=ConstrainColormapIndex(image,(*p >> 2) & 0x3,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
-            q++;
-            index=ConstrainColormapIndex(image,(*p) & 0x3);
-            SetPixelIndex(indexes+x+1,index);
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+            q+=GetPixelChannels(image);
+            index=ConstrainColormapIndex(image,(*p) & 0x3,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
-            q++;
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+            q+=GetPixelChannels(image);
             p++;
         }
        if ((image->columns % 4) != 0)
           {
-            index=ConstrainColormapIndex(image,(*p >> 6) & 0x3);
-            SetPixelIndex(indexes+x,index);
+            index=ConstrainColormapIndex(image,(*p >> 6) & 0x3,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
-            q++;
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+            q+=GetPixelChannels(image);
             if ((image->columns % 4) > 1)
               {
-                index=ConstrainColormapIndex(image,(*p >> 4) & 0x3);
-                SetPixelIndex(indexes+x,index);
+                index=ConstrainColormapIndex(image,(*p >> 4) & 0x3,exception);
+                SetPixelIndex(image,index,q);
                 if (index < image->colors)
-                  SetPixelRGBO(q,image->colormap+(ssize_t) index);
-                q++;
+                  SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+                q+=GetPixelChannels(image);
                 if ((image->columns % 4) > 2)
                   {
-                    index=ConstrainColormapIndex(image,(*p >> 2) & 0x3);
-                    SetPixelIndex(indexes+x,index);
+                    index=ConstrainColormapIndex(image,(*p >> 2) & 0x3,
+                      exception);
+                    SetPixelIndex(image,index,q);
                     if (index < image->colors)
-                      SetPixelRGBO(q,image->colormap+(ssize_t) index);
-                    q++;
+                      SetPixelViaPixelInfo(image,image->colormap+(ssize_t)
+                        index,q);
+                    q+=GetPixelChannels(image);
                   }
               }
             p++;
@@ -381,26 +372,26 @@ static MagickBooleanType InsertRow(unsigned char *p,ssize_t y,Image *image,
       {
         for (x=0; x < ((ssize_t) image->columns-1); x+=2)
           {
-            index=ConstrainColormapIndex(image,(*p >> 4) & 0x0f);
-            SetPixelIndex(indexes+x,index);
+            index=ConstrainColormapIndex(image,(*p >> 4) & 0x0f,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
-            q++;
-            index=ConstrainColormapIndex(image,(*p) & 0x0f);
-            SetPixelIndex(indexes+x+1,index);
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
+            q+=GetPixelChannels(image);
+            index=ConstrainColormapIndex(image,(*p) & 0x0f,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
             p++;
-            q++;
+            q+=GetPixelChannels(image);
           }
         if ((image->columns % 2) != 0)
           {
-            index=ConstrainColormapIndex(image,(*p >> 4) & 0x0f);
-            SetPixelIndex(indexes+x,index);
+            index=ConstrainColormapIndex(image,(*p >> 4) & 0x0f,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
             p++;
-            q++;
+            q+=GetPixelChannels(image);
           }
         break;
       }
@@ -408,12 +399,12 @@ static MagickBooleanType InsertRow(unsigned char *p,ssize_t y,Image *image,
       {
         for (x=0; x < (ssize_t) image->columns; x++)
           {
-            index=ConstrainColormapIndex(image,*p);
-            SetPixelIndex(indexes+x,index);
+            index=ConstrainColormapIndex(image,*p,exception);
+            SetPixelIndex(image,index,q);
             if (index < image->colors)
-              SetPixelRGBO(q,image->colormap+(ssize_t) index);
+              SetPixelViaPixelInfo(image,image->colormap+(ssize_t) index,q);
             p++;
-            q++;
+            q+=GetPixelChannels(image);
           }
       }
       break;
@@ -421,10 +412,10 @@ static MagickBooleanType InsertRow(unsigned char *p,ssize_t y,Image *image,
     case 24:     /*  Convert DirectColor scanline.  */
       for (x=0; x < (ssize_t) image->columns; x++)
         {
-          SetPixelRed(q,ScaleCharToQuantum(*p++));
-          SetPixelGreen(q,ScaleCharToQuantum(*p++));
-          SetPixelBlue(q,ScaleCharToQuantum(*p++));
-          q++;
+          SetPixelRed(image,ScaleCharToQuantum(*p++),q);
+          SetPixelGreen(image,ScaleCharToQuantum(*p++),q);
+          SetPixelBlue(image,ScaleCharToQuantum(*p++),q);
+          q+=GetPixelChannels(image);
         }
       break;
     }
@@ -441,13 +432,13 @@ static MagickBooleanType InsertRow(unsigned char *p,ssize_t y,Image *image,
   x++; \
   if((ssize_t) x>=ldblk) \
   { \
-    if (InsertRow(BImgBuff,(ssize_t) y,image,bpp) != MagickFalse) \
+    if (InsertRow(image,BImgBuff,(ssize_t) y,bpp,exception) != MagickFalse) \
       y++; \
     x=0; \
   } \
 }
 /* WPG1 raster reader. */
-static int UnpackWPGRaster(Image *image,int bpp)
+static int UnpackWPGRaster(Image *image,int bpp,ExceptionInfo *exception)
 {
   int
     x,
@@ -472,14 +463,14 @@ static int UnpackWPGRaster(Image *image,int bpp)
   (void) memset(BImgBuff,0,(size_t) ldblk*8*sizeof(*BImgBuff));
 
   while(y<(ssize_t) image->rows)
-  {
-    int
-      c;
+    {
+      int
+        c;
 
-    c=ReadBlobByte(image);
-    if (c ==  EOF)
-      break;
-    bbuf=(unsigned char) c;
+      c=ReadBlobByte(image);
+      if (c == EOF)
+        break;
+      bbuf=(unsigned char) c;
       RunCount=bbuf & 0x7F;
       if(bbuf & 0x80)
         {
@@ -530,7 +521,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
                   BImgBuff=(unsigned char *) RelinquishMagickMemory(BImgBuff);
                   return(-4);
                 }
-              if (InsertRow(BImgBuff,y-1,image,bpp) == MagickFalse)
+              if (InsertRow(image,BImgBuff,y-1,bpp,exception) == MagickFalse)
                 {
                   BImgBuff=(unsigned char *) RelinquishMagickMemory(BImgBuff);
                   return(-5);
@@ -542,7 +533,7 @@ static int UnpackWPGRaster(Image *image,int bpp)
         break;
     }
   BImgBuff=(unsigned char *) RelinquishMagickMemory(BImgBuff);
-  return(y < (ssize_t) image->rows ? -5 : 0);
+  return(y <(ssize_t) image->rows ? -5 : 0);
 }
 
 
@@ -558,25 +549,26 @@ RestoreMSCWarning \
   x++; \
   if((ssize_t) x >= ldblk) \
   { \
-    if (InsertRow(BImgBuff,(ssize_t) y,image,bpp) != MagickFalse) \
+    if (InsertRow(image,BImgBuff,(ssize_t) y,bpp,exception) != MagickFalse) \
       y++; \
     x=0; \
-  } \
+   } \
 }
 /* WPG2 raster reader. */
-static int UnpackWPG2Raster(Image *image,int bpp)
+static int UnpackWPG2Raster(Image *image,int bpp,ExceptionInfo *exception)
 {
-  int XorMe = 0;
-
   int
-    RunCount;
+    RunCount,
+    XorMe = 0;
+
+  register ssize_t
+    i;
 
   size_t
     x,
     y;
 
   ssize_t
-    i,
     ldblk;
 
   unsigned int
@@ -625,7 +617,7 @@ static int UnpackWPG2Raster(Image *image,int bpp)
           RunCount=ReadBlobByte(image);   /* BLK */
           if (RunCount < 0)
             break;
-          for(i=0; i < SampleSize*(RunCount+1); i++)
+          for(i=0; i < ((ssize_t) SampleSize*(RunCount+1)); i++)
             {
               InsertByte6(0);
             }
@@ -654,8 +646,12 @@ static int UnpackWPG2Raster(Image *image,int bpp)
             /* duplicate the previous row RunCount x */
             for(i=0;i<=RunCount;i++)
               {
-                if (InsertRow(BImgBuff,(ssize_t) (image->rows > y ? y : image->rows-1),image,bpp) != MagickFalse)
-                  y++;
+                if (InsertRow(image,BImgBuff,(ssize_t) (image->rows > y ? y : image->rows-1),bpp,exception) == MagickFalse)
+                  {
+                    BImgBuff=(unsigned char *) RelinquishMagickMemory(BImgBuff);
+                    return(-3);
+                  }
+                y++;
               }
           }
           break;
@@ -663,10 +659,10 @@ static int UnpackWPG2Raster(Image *image,int bpp)
           RunCount=ReadBlobByte(image);   /* WHT */
           if (RunCount < 0)
             break;
-          for (i=0; i < SampleSize*(RunCount+1); i++)
-          {
-            InsertByte6(0xFF);
-          }
+          for(i=0; i < ((ssize_t) SampleSize*(RunCount+1)); i++)
+            {
+              InsertByte6(0xFF);
+            }
           break;
         default:
           RunCount=bbuf & 0x7F;
@@ -680,7 +676,7 @@ static int UnpackWPG2Raster(Image *image,int bpp)
                   InsertByte6(SampleBuffer[bbuf]);
             }
           else {      /* NRP */
-            for(i=0; i< SampleSize*(RunCount+1);i++)
+            for(i=0; i < (ssize_t) (SampleSize*(RunCount+1)); i++)
               {
                 bbuf=ReadBlobByte(image);
                 InsertByte6(bbuf);
@@ -762,7 +758,7 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
   MagickOffsetType PS_Offset,ssize_t PS_Size,ExceptionInfo *exception)
 {
   char
-    postscript_file[MaxTextExtent];
+    postscript_file[MagickPathExtent];
 
   const MagicInfo
     *magic_info;
@@ -783,7 +779,7 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
     status;
 
   unsigned char
-    magick[2*MaxTextExtent];
+    magick[2*MagickPathExtent];
 
   ssize_t
     count;
@@ -808,7 +804,7 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
         image->filename);
       goto FINISH_UNL;
     }
-  count=ReadBlob(image, 2*MaxTextExtent, magick);
+  count=ReadBlob(image, 2*MagickPathExtent, magick);
   if (count < 1)
     {
       (void) fclose(ps_file);
@@ -828,7 +824,7 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
   {
     c=ReadBlobByte(image);
     if (c == EOF)
-      {
+      {      
         (void) fclose(ps_file);
         ThrowException(exception,CorruptImageError,"ImproperImageHeader",
           image->filename);
@@ -843,10 +839,10 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
   if(magic_info == (const MagicInfo *) NULL) goto FINISH_UNL;
   /*     printf("Detected:%s  \n",magic_info->name); */
   if(exception->severity != UndefinedException) goto FINISH_UNL;
-  if(magic_info->name == (char *) NULL) goto FINISH_UNL;
-  (void) strncpy(clone_info->magick,magic_info->name,MaxTextExtent-1);
+  (void) CopyMagickString(clone_info->magick,GetMagicName(magic_info),
+    MagickPathExtent);
   if (LocaleCompare(clone_info->magick,"PFB") != 0)
-    {
+    {      
       ThrowException(exception,CorruptImageError,"ImproperImageHeader",
         image->filename);
       goto FINISH_UNL;
@@ -885,10 +881,10 @@ static Image *ExtractPostscript(Image *image,const ImageInfo *image_info,
         {
           DeleteImageFromList(&p);
           if (p == (Image *) NULL)
-          {
-            image2=(Image *) NULL;
-            goto FINISH_UNL;
-          }
+            {
+              image2=(Image *) NULL;
+              goto FINISH_UNL;
+            }
         }
       else
         {
@@ -1083,7 +1079,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickCoreSignature);
   one=1;
-  image=AcquireImage(image_info);
+  image=AcquireImage(image_info,exception);
   image->depth=8;
   status=OpenBlob(image_info,image,ReadBinaryBlobMode,exception);
   if (status == MagickFalse)
@@ -1115,7 +1111,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
   (void) ResetImagePixels(image,exception);
   bpp=0;
   BitmapHeader2.RotAngle=0;
-  Rec2.RecordLength = 0;
+  Rec2.RecordLength=0;
 
   switch(Header.FileType)
     {
@@ -1152,8 +1148,8 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               if(BitmapHeader1.HorzRes && BitmapHeader1.VertRes)
                 {
                   image->units=PixelsPerCentimeterResolution;
-                  image->x_resolution=BitmapHeader1.HorzRes/470.0;
-                  image->y_resolution=BitmapHeader1.VertRes/470.0;
+                  image->resolution.x=BitmapHeader1.HorzRes/470.0;
+                  image->resolution.y=BitmapHeader1.VertRes/470.0;
                 }
               image->columns=BitmapHeader1.Width;
               image->rows=BitmapHeader1.Height;
@@ -1165,12 +1161,12 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               WPG_Palette.StartIndex=ReadBlobLSBShort(image);
               WPG_Palette.NumOfEntries=ReadBlobLSBShort(image);
               if ((WPG_Palette.NumOfEntries-WPG_Palette.StartIndex) >
-                  (Rec2.RecordLength-2-2) / 3)
+                  (Rec2.RecordLength-2-2)/3)
                 ThrowReaderException(CorruptImageError,"InvalidColormapIndex");
               if (WPG_Palette.StartIndex > WPG_Palette.NumOfEntries)
                 ThrowReaderException(CorruptImageError,"InvalidColormapIndex");
               image->colors=WPG_Palette.NumOfEntries;
-              if (AcquireImageColormap(image,image->colors) == MagickFalse)
+              if (AcquireImageColormap(image,image->colors,exception) == MagickFalse)
                 goto NoMemory;
               for (i=WPG_Palette.StartIndex;
                    i < (int)WPG_Palette.NumOfEntries; i++)
@@ -1181,7 +1177,6 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                     ReadBlobByte(image));
                   image->colormap[i].blue=ScaleCharToQuantum((unsigned char)
                     ReadBlobByte(image));
-                  image->colormap[i].opacity=OpaqueOpacity;
                 }
               break;
 
@@ -1220,15 +1215,15 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               image->page.y=(int) (BitmapHeader2.LowLeftX/470.0);
               if(BitmapHeader2.HorzRes && BitmapHeader2.VertRes)
                 {
-                  image->x_resolution=BitmapHeader2.HorzRes/470.0;
-                  image->y_resolution=BitmapHeader2.VertRes/470.0;
+                  image->resolution.x=BitmapHeader2.HorzRes/470.0;
+                  image->resolution.y=BitmapHeader2.VertRes/470.0;
                 }
               image->columns=BitmapHeader2.Width;
               image->rows=BitmapHeader2.Height;
               bpp=BitmapHeader2.Depth;
 
             UnpackRaster:
-              status=SetImageExtent(image,image->columns,image->rows);
+              status=SetImageExtent(image,image->columns,image->rows,exception);
               if (status == MagickFalse)
                 break;
               (void) ResetImagePixels(image,exception);
@@ -1238,7 +1233,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                   if (image->colors > GetBlobSize(image))
                     ThrowReaderException(CorruptImageError,
                       "InsufficientImageDataInFile");
-                  if (!AcquireImageColormap(image,image->colors))
+                  if (!AcquireImageColormap(image,image->colors,exception))
                     {
                     NoMemory:
                       ThrowReaderException(ResourceLimitError,
@@ -1250,7 +1245,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                       image->colormap[i].red=ScaleCharToQuantum(WPG1_Palette[i].Red);
                       image->colormap[i].green=ScaleCharToQuantum(WPG1_Palette[i].Green);
                       image->colormap[i].blue=ScaleCharToQuantum(WPG1_Palette[i].Blue);
-                      image->colormap[i].opacity=OpaqueOpacity;
+                      image->colormap[i].alpha=OpaqueAlpha;
                     }
                 }
               else
@@ -1258,7 +1253,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                   if (bpp < 24)
                   if ( (image->colors < (one << bpp)) && (bpp != 24) )
                     {
-                      PixelPacket
+                      PixelInfo
                         *colormap;
 
                       size_t
@@ -1266,16 +1261,16 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
 
                       colormap=image->colormap;
                       colors=image->colors;
-                      image->colormap=(PixelPacket *) NULL;
-                      if (AcquireImageColormap(image,one << bpp) == MagickFalse)
+                      image->colormap=(PixelInfo *) NULL;
+                      if (AcquireImageColormap(image,one << bpp,exception) == MagickFalse)
                         {
-                          colormap=(PixelPacket *)
+                          colormap=(PixelInfo *)
                             RelinquishMagickMemory(colormap);
                           goto NoMemory;
                         }
-                      (void) memcpy(image->colormap,colormap,colors*
-                        sizeof(*image->colormap));
-                      colormap=(PixelPacket *)
+                      (void) memcpy(image->colormap,colormap,MagickMin(
+                        image->colors,colors)*sizeof(*image->colormap));
+                      colormap=(PixelInfo *)
                         RelinquishMagickMemory(colormap);
                     }
                 }
@@ -1292,17 +1287,16 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                       image->colormap[1].red =
                         image->colormap[1].green =
                         image->colormap[1].blue = QuantumRange;
-                      image->colormap[1].opacity=OpaqueOpacity;
+                      image->colormap[1].alpha=OpaqueAlpha;
                     }
                 }
-
               if(!image_info->ping)
-                if(UnpackWPGRaster(image,bpp) < 0)
+                if(UnpackWPGRaster(image,bpp,exception) < 0)
                   /* The raster cannot be unpacked */
                   {
                   DecompressionFailed:
                     ThrowReaderException(CoderError,"UnableToDecompressImage");
-                      }
+                  }
 
               if(Rec.RecType==0x14 && BitmapHeader2.RotAngle!=0 && !image_info->ping)
                 {
@@ -1346,7 +1340,11 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                 }
 
               /* Allocate next image structure. */
-              AcquireNextImage(image_info,image);
+              if ((image_info->ping != MagickFalse) &&
+                  (image_info->number_scenes != 0))
+                if (image->scene >= (image_info->scene+image_info->number_scenes-1))
+                  goto Finish;
+              AcquireNextImage(image_info,image,exception);
               image->depth=8;
               if (image->next == (Image *) NULL)
                 goto Finish;
@@ -1356,7 +1354,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               break;
 
             case 0x1B:  /* Postscript l2 */
-              if (Rec.RecordLength > 0x3C)
+              if (Rec.RecordLength>0x3C)
                 {
                   image=ExtractPostscript(image,image_info,
                     TellBlob(image)+0x3C,   /* skip PS l2 header in the wpg */
@@ -1377,7 +1375,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
         {
           if (SeekBlob(image,Header.DataOffset,SEEK_SET) != Header.DataOffset)
             break;
-          if(EOFBlob(image))
+          if (EOFBlob(image))
             break;
 
           Rec2.Class=(i=ReadBlobByte(image));
@@ -1409,7 +1407,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               if (WPG_Palette.StartIndex >= WPG_Palette.NumOfEntries)
                 ThrowReaderException(CorruptImageError,"InvalidColormapIndex");
               image->colors=WPG_Palette.NumOfEntries;
-              if (AcquireImageColormap(image,image->colors) == MagickFalse)
+              if (AcquireImageColormap(image,image->colors,exception) == MagickFalse)
                 ThrowReaderException(ResourceLimitError,
                   "MemoryAllocationFailed");
               for (i=WPG_Palette.StartIndex;
@@ -1421,7 +1419,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                     ReadBlobByte(image));
                   image->colormap[i].blue=ScaleCharToQuantum((char)
                     ReadBlobByte(image));
-                  image->colormap[i].opacity=OpaqueOpacity;
+                  image->colormap[i].alpha=OpaqueAlpha;
                   (void) ReadBlobByte(image);   /*Opacity??*/
                 }
               break;
@@ -1459,26 +1457,22 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
               image->rows=Bitmap2Header1.Height;
               if (image_info->ping != MagickFalse)
                 return(image);
-              status=SetImageExtent(image,image->columns,image->rows);
+              status=SetImageExtent(image,image->columns,image->rows,exception);
               if (status != MagickFalse)
                 status=ResetImagePixels(image,exception);
               if (status == MagickFalse)
                 break;
               if ((image->colors == 0) && (bpp != 24))
                 {
-                  size_t
-                    one;
-
-                  one=1;
                   image->colors=one << bpp;
-                  if (!AcquireImageColormap(image,image->colors))
+                  if (!AcquireImageColormap(image,image->colors,exception))
                     goto NoMemory;
                 }
               else
                 {
                   if(bpp < 24)
                     if( image->colors<(one << bpp) && bpp!=24 )
-                      image->colormap=(PixelPacket *) ResizeQuantumMemory(
+                      image->colormap=(PixelInfo *) ResizeQuantumMemory(
                        image->colormap,(size_t) (one << bpp),
                        sizeof(*image->colormap));
                 }
@@ -1493,7 +1487,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                       ldblk+1,sizeof(*BImgBuff));
                     if (BImgBuff == (unsigned char *) NULL)
                       goto NoMemory;
-                    for (i=0; i < (ssize_t) image->rows; i++)
+                    for (i=0; i< (ssize_t) image->rows; i++)
                     {
                       ssize_t
                         count;
@@ -1501,7 +1495,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                       count=ReadBlob(image,(size_t) ldblk,BImgBuff);
                       if (count != ldblk)
                         break;
-                      if (InsertRow(BImgBuff,i,image,bpp) == MagickFalse)
+                      if (InsertRow(image,BImgBuff,i,bpp,exception) == MagickFalse)
                         break;
                     }
                     BImgBuff=(unsigned char *) RelinquishMagickMemory(BImgBuff);
@@ -1511,7 +1505,7 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                   }
                 case 1:    /*RLE for WPG2 */
                   {
-                    if( UnpackWPG2Raster(image,bpp) < 0)
+                    if( UnpackWPG2Raster(image,bpp,exception) < 0)
                       goto DecompressionFailed;
                     break;
                   }
@@ -1538,22 +1532,26 @@ static Image *ReadWPGImage(const ImageInfo *image_info,
                   Image
                     *flip_image;
 
-                  flip_image = FlipImage(image, exception);
-                  if (flip_image != (Image *) NULL) {
-                    DuplicateBlob(flip_image,image);
-                    ReplaceImageInList(&image,flip_image);
-                  }
+                   flip_image = FlipImage(image, exception);
+                   if (flip_image != (Image *) NULL) {
+                     DuplicateBlob(flip_image,image);
+                     ReplaceImageInList(&image,flip_image);
+                    }
                   /* Try to change CTM according to Flip - I am not sure, must be checked.
                      float_matrix Tx(3,3);
                      Tx(0,0)= 1;   Tx(1,0)= 0;   Tx(2,0)=0;
                      Tx(0,1)= 0;   Tx(1,1)=-1;   Tx(2,1)=0;
                      Tx(0,2)= 0;   Tx(1,2)=(WPG._2Rect.Y_ur+WPG._2Rect.Y_ll);
                      Tx(2,2)=1; */
-                }
+              }
 
 
               /* Allocate next image structure. */
-              AcquireNextImage(image_info,image);
+              if ((image_info->ping != MagickFalse) &&
+                  (image_info->number_scenes != 0))
+                if (image->scene >= (image_info->scene+image_info->number_scenes-1))
+                  goto Finish;
+              AcquireNextImage(image_info,image,exception);
               image->depth=8;
               if (image->next == (Image *) NULL)
                 goto Finish;
@@ -1656,12 +1654,10 @@ ModuleExport size_t RegisterWPGImage(void)
   MagickInfo
     *entry;
 
-  entry=SetMagickInfo("WPG");
+  entry=AcquireMagickInfo("WPG","WPG","Word Perfect Graphics");
   entry->decoder=(DecodeImageHandler *) ReadWPGImage;
   entry->magick=(IsImageFormatHandler *) IsWPG;
-  entry->description=AcquireString("Word Perfect Graphics");
-  entry->module=ConstantString("WPG");
-  entry->seekable_stream=MagickTrue;
+  entry->flags|=CoderDecoderSeekableStreamFlag;
   (void) RegisterMagickInfo(entry);
   return(MagickImageCoderSignature);
 }
